from __future__ import annotations

import base64
import json
import logging
import os
import struct
import time
from dataclasses import asdict, dataclass
from typing import Any, Callable, Literal, cast
from urllib.parse import urlparse
from urllib.request import urlopen

import openai
from openai.types.chat import (
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionMessageParam,
    ChatCompletionUserMessageParam,
)

logger = logging.getLogger(__name__)

BASE_URL = os.environ.get("CODEX_BASE_URL", "http://localhost:2455/v1")
API_KEY = os.environ.get("CODEX_API_KEY", "sk-local")
MODEL_OVERRIDE = os.environ.get("CODEX_MODEL")

FILE_URL = os.environ.get(
    "CODEX_TEST_FILE_URL",
    "https://www.berkshirehathaway.com/letters/2024ltr.pdf",
)
FILE_ID = os.environ.get("CODEX_TEST_FILE_ID", "file-unknown")
VECTOR_STORE_ID = os.environ.get("CODEX_VECTOR_STORE_ID")
IMAGE_URL = os.environ.get(
    "CODEX_TEST_IMAGE_URL",
    "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/"
    "Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-"
    "madison-the-nature-boardwalk.jpg",
)
AUDIO_B64 = os.environ.get("CODEX_TEST_AUDIO_B64")
AUDIO_FORMAT = os.environ.get("CODEX_TEST_AUDIO_FORMAT", "wav")
CHAT_FILE_ID = os.environ.get("CODEX_TEST_CHAT_FILE_ID", "file-unknown")
CHAT_FILE_DATA_B64 = os.environ.get("CODEX_TEST_CHAT_FILE_DATA_B64")
CHAT_FILE_NAME = os.environ.get("CODEX_TEST_CHAT_FILE_NAME", "input.bin")

RUN_RESPONSES = os.environ.get("CODEX_RUN_RESPONSES", "1") != "0"
RUN_CHAT = os.environ.get("CODEX_RUN_CHAT", "1") != "0"
RUN_TOOLS = os.environ.get("CODEX_RUN_TOOLS", "1") != "0"
RUN_REASONING = os.environ.get("CODEX_RUN_REASONING", "1") != "0"

DEFAULT_TEXT = "Return 'ok' only."

EXPECTED_UNSUPPORTED = [
    "R.file_input_id",
    "R.previous_response_id",
    "R.truncation_auto",
    "T.file_search",
    "T.code_interpreter",
    "T.computer_use",
    "T.image_generation",
    "T.tool_choice_required",
    "C.audio_input",
    "C.file_input_id",
]


class CaseSkipped(Exception):
    def __init__(self, reason: str) -> None:
        super().__init__(reason)
        self.reason = reason


class CaseError(Exception):
    def __init__(self, detail: dict[str, Any]) -> None:
        super().__init__(detail.get("message", "case error"))
        self.detail = detail


@dataclass
class CaseResult:
    name: str
    status: str
    detail: dict[str, Any]


def _error_detail(exc: Exception) -> dict[str, Any]:
    detail: dict[str, Any] = {
        "type": exc.__class__.__name__,
        "message": str(exc),
    }
    status_code = getattr(exc, "status_code", None)
    if status_code is not None:
        detail["status_code"] = status_code
    body = getattr(exc, "body", None)
    if body is not None:
        detail["body"] = body
    response = getattr(exc, "response", None)
    if response is not None:
        try:
            detail["response_json"] = response.json()
        except (ValueError, TypeError, AttributeError):
            try:
                detail["response_text"] = response.text
            except (AttributeError, TypeError):
                pass
    return detail


def _chat_user_message(content: str | list[ChatCompletionContentPartParam]) -> ChatCompletionMessageParam:
    return ChatCompletionUserMessageParam(role="user", content=content)


def _chat_text_part(text: str) -> ChatCompletionContentPartTextParam:
    return {"type": "text", "text": text}


def _chat_image_part(url: str) -> ChatCompletionContentPartParam:
    return {"type": "image_url", "image_url": {"url": url}}


def _chat_audio_format(value: str) -> Literal["wav", "mp3"]:
    normalized = value.strip().lower()
    if normalized in ("wav", "mp3"):
        return cast(Literal["wav", "mp3"], normalized)
    raise CaseSkipped("C.audio_input requires CODEX_TEST_AUDIO_FORMAT to be wav or mp3")


def _chat_audio_part(data_b64: str, audio_format: Literal["wav", "mp3"]) -> ChatCompletionContentPartInputAudioParam:
    return {"type": "input_audio", "input_audio": {"data": data_b64, "format": audio_format}}


def _chat_file_part_from_data(*, file_data_b64: str, filename: str) -> ChatCompletionContentPartParam:
    return {"type": "file", "file": {"file_data": file_data_b64, "filename": filename}}


def _chat_file_data_from_url(file_url: str) -> tuple[str, str]:
    parsed = urlparse(file_url)
    filename = parsed.path.rsplit("/", 1)[-1] or "input.bin"
    with urlopen(file_url, timeout=15) as response:  # nosec B310
        payload = response.read()
    return base64.b64encode(payload).decode("ascii"), filename


def run_case(name: str, fn: Callable[[], Any]) -> CaseResult:
    try:
        result = fn()
        return CaseResult(name=name, status="ok", detail={"result": result})
    except CaseSkipped as exc:
        return CaseResult(name=name, status="skipped", detail={"reason": exc.reason})
    except CaseError as exc:
        return CaseResult(name=name, status="error", detail=exc.detail)
    except Exception as exc:
        return CaseResult(name=name, status="error", detail=_error_detail(exc))


def _pick_model(client: openai.OpenAI) -> str:
    if MODEL_OVERRIDE:
        return MODEL_OVERRIDE
    models = client.models.list()
    if models.data:
        return models.data[0].id
    raise RuntimeError("No models returned from /v1/models")


def _input_file_payload(file_ref: dict[str, Any]) -> list[dict[str, Any]]:
    return [
        {
            "role": "user",
            "content": [
                {"type": "input_text", "text": "Summarize this file in one sentence."},
                {"type": "input_file", **file_ref},
            ],
        }
    ]


def _input_text_payload(text: str) -> list[dict[str, Any]]:
    return [
        {
            "role": "user",
            "content": [{"type": "input_text", "text": text}],
        }
    ]


def _build_wav_base64(duration_ms: int = 100, sample_rate: int = 8000) -> str:
    num_channels = 1
    sample_width = 2
    num_frames = int(sample_rate * duration_ms / 1000)
    data = b"\x00\x00" * num_frames
    byte_rate = sample_rate * num_channels * sample_width
    block_align = num_channels * sample_width
    riff_size = 36 + len(data)
    header = b"RIFF" + struct.pack("<I", riff_size) + b"WAVE"
    fmt = b"fmt " + struct.pack(
        "<IHHIIHH",
        16,
        1,
        num_channels,
        sample_rate,
        byte_rate,
        block_align,
        sample_width * 8,
    )
    data_chunk = b"data" + struct.pack("<I", len(data)) + data
    wav_bytes = header + fmt + data_chunk
    return base64.b64encode(wav_bytes).decode("ascii")


def _response_stream_summary(stream: Any) -> dict[str, Any]:
    event_types: list[str] = []
    output_text = ""
    for event in stream:
        event_type = getattr(event, "type", None)
        if isinstance(event_type, str):
            event_types.append(event_type)
        if event_type == "response.output_text.delta":
            delta = getattr(event, "delta", "")
            if isinstance(delta, str):
                output_text += delta
    return {
        "event_types": event_types,
        "event_count": len(event_types),
        "output_text": output_text,
    }


def _chat_stream_summary(stream: Any) -> dict[str, Any]:
    output_text = ""
    chunk_count = 0
    for chunk in stream:
        chunk_count += 1
        if not chunk.choices:
            continue
        delta = chunk.choices[0].delta
        if delta and delta.content:
            output_text += delta.content
    return {"chunk_count": chunk_count, "output_text": output_text}


def main() -> int:
    logging.basicConfig(level=logging.INFO, format="%(message)s", force=True)
    client = openai.OpenAI(api_key=API_KEY, base_url=BASE_URL)
    try:
        model = _pick_model(client)
    except (openai.APIError, RuntimeError) as exc:
        logger.error("Failed to list models: %s", exc)
        return 2

    audio_b64 = AUDIO_B64 or _build_wav_base64()

    logger.info("Base URL: %s", BASE_URL)
    logger.info("Model: %s", model)
    logger.info("File URL: %s", FILE_URL)
    logger.info("File ID: %s", FILE_ID)
    logger.info("Vector Store ID: %s", VECTOR_STORE_ID or "(missing)")
    logger.info("Image URL: %s", IMAGE_URL)
    logger.info("Audio format: %s", AUDIO_FORMAT)
    logger.info("Chat file ID: %s", CHAT_FILE_ID)
    logger.info("Expected unsupported: %s", ", ".join(EXPECTED_UNSUPPORTED))

    results: list[CaseResult] = []

    cases: list[tuple[str, Callable[[], Any]]] = []

    if RUN_RESPONSES:

        def r_text_input_string() -> dict[str, Any]:
            resp = client.responses.create(model=model, input=DEFAULT_TEXT)
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def r_text_input_list() -> dict[str, Any]:
            resp = client.responses.create(model=model, input=cast(Any, _input_text_payload(DEFAULT_TEXT)))
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def r_image_input_url() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(
                    Any,
                    [
                        {
                            "role": "user",
                            "content": [
                                {"type": "input_text", "text": "Describe the image."},
                                {"type": "input_image", "image_url": IMAGE_URL},
                            ],
                        }
                    ],
                ),
            )
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def r_file_input_url() -> dict[str, Any]:
            resp = client.responses.create(model=model, input=cast(Any, _input_file_payload({"file_url": FILE_URL})))
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def r_file_input_id() -> dict[str, Any]:
            resp = client.responses.create(model=model, input=cast(Any, _input_file_payload({"file_id": FILE_ID})))
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def r_streaming() -> dict[str, Any]:
            stream = client.responses.create(
                model=model, input=cast(Any, _input_text_payload("Stream ok")), stream=True
            )
            return _response_stream_summary(stream)

        def r_previous_response_id() -> dict[str, Any]:
            first = client.responses.create(model=model, input=cast(Any, _input_text_payload(DEFAULT_TEXT)))
            first_id = getattr(first, "id", None)
            if not first_id:
                raise CaseError({"message": "missing response id from first call"})
            try:
                second = client.responses.create(
                    model=model,
                    input=cast(Any, _input_text_payload("Repeat exactly: ok")),
                    previous_response_id=first_id,
                )
            except Exception as exc:
                detail = _error_detail(exc)
                detail["first_id"] = first_id
                raise CaseError(detail) from exc
            return {
                "first_id": first_id,
                "second_id": getattr(second, "id", None),
                "second_output_text": getattr(second, "output_text", None),
            }

        def r_truncation_auto() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Short reply.")),
                truncation="auto",
            )
            return {"response_id": getattr(resp, "id", None)}

        def r_text_format_json_object() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload('Return {"ok": true} as JSON.')),
                text={"format": {"type": "json_object"}},
            )
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def r_include_logprobs() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Return ok.")),
                include=["message.output_text.logprobs"],
            )
            return {"response_id": getattr(resp, "id", None)}

        cases.extend(
            [
                ("R.text_input_string", r_text_input_string),
                ("R.text_input_list", r_text_input_list),
                ("R.image_input_url", r_image_input_url),
                ("R.file_input_url", r_file_input_url),
                ("R.file_input_id", r_file_input_id),
                ("R.streaming", r_streaming),
                ("R.previous_response_id", r_previous_response_id),
                ("R.truncation_auto", r_truncation_auto),
                ("R.text_format_json_object", r_text_format_json_object),
                ("R.include_logprobs", r_include_logprobs),
            ]
        )

    if RUN_REASONING:

        def r_reasoning_effort_low() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Solve 1+1.")),
                reasoning={"effort": "low"},
            )
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def r_reasoning_summary_detailed() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Explain briefly.")),
                reasoning={"summary": "detailed"},
            )
            return {"response_id": getattr(resp, "id", None)}

        def r_reasoning_encrypted_content() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Reason about 2+2.")),
                include=["reasoning.encrypted_content"],
            )
            return {"response_id": getattr(resp, "id", None)}

        cases.extend(
            [
                ("R.reasoning_effort_low", r_reasoning_effort_low),
                ("R.reasoning_summary_detailed", r_reasoning_summary_detailed),
                ("R.reasoning_encrypted_content", r_reasoning_encrypted_content),
            ]
        )

    if RUN_TOOLS:

        def t_web_search() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("What was a positive news story today?")),
                tools=[{"type": "web_search"}],
            )
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def t_file_search() -> dict[str, Any]:
            vector_store_id = VECTOR_STORE_ID or "vs_dummy"
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Find notes about onboarding.")),
                tools=[{"type": "file_search", "vector_store_ids": [vector_store_id]}],
                include=["file_search_call.results"],
            )
            return {
                "response_id": getattr(resp, "id", None),
                "output_text": getattr(resp, "output_text", None),
                "vector_store_id": vector_store_id,
                "used_dummy_vector_store_id": VECTOR_STORE_ID is None,
            }

        def t_code_interpreter() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Compute 2+2 and return only the number.")),
                tools=[{"type": "code_interpreter", "container": {"type": "auto"}}],
            )
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def t_computer_use() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Open example.com and report the page title.")),
                tools=[
                    {
                        "type": "computer_use_preview",
                        "display_width": 1024,
                        "display_height": 768,
                        "environment": "browser",
                    }
                ],
            )
            return {"response_id": getattr(resp, "id", None), "output_text": getattr(resp, "output_text", None)}

        def t_image_generation() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Generate a small red circle on white background.")),
                tools=[{"type": "image_generation"}],
            )
            return {"response_id": getattr(resp, "id", None)}

        def t_tool_choice_required() -> dict[str, Any]:
            resp = client.responses.create(
                model=model,
                input=cast(Any, _input_text_payload("Draw a red dot.")),
                tools=[{"type": "image_generation"}],
                tool_choice="required",
            )
            return {"response_id": getattr(resp, "id", None)}

        cases.extend(
            [
                ("T.web_search", t_web_search),
                ("T.file_search", t_file_search),
                ("T.code_interpreter", t_code_interpreter),
                ("T.computer_use", t_computer_use),
                ("T.image_generation", t_image_generation),
                ("T.tool_choice_required", t_tool_choice_required),
            ]
        )

    if RUN_CHAT:

        def c_text() -> dict[str, Any]:
            messages: list[ChatCompletionMessageParam] = [
                _chat_user_message(DEFAULT_TEXT),
            ]
            resp = client.chat.completions.create(
                model=model,
                messages=messages,
            )
            return {"id": getattr(resp, "id", None), "content": resp.choices[0].message.content}

        def c_image_url() -> dict[str, Any]:
            messages: list[ChatCompletionMessageParam] = [
                _chat_user_message(
                    [
                        _chat_text_part("Describe the image."),
                        _chat_image_part(IMAGE_URL),
                    ]
                )
            ]
            resp = client.chat.completions.create(
                model=model,
                messages=messages,
            )
            return {"id": getattr(resp, "id", None), "content": resp.choices[0].message.content}

        def c_audio_input() -> dict[str, Any]:
            audio_format = _chat_audio_format(AUDIO_FORMAT)
            messages: list[ChatCompletionMessageParam] = [
                _chat_user_message(
                    [
                        _chat_text_part("Transcribe the audio."),
                        _chat_audio_part(audio_b64, audio_format),
                    ]
                )
            ]
            resp = client.chat.completions.create(
                model=model,
                messages=messages,
            )
            return {"id": getattr(resp, "id", None), "content": resp.choices[0].message.content}

        def c_file_input() -> dict[str, Any]:
            try:
                if CHAT_FILE_DATA_B64:
                    file_data_b64 = CHAT_FILE_DATA_B64
                    filename = CHAT_FILE_NAME
                else:
                    file_data_b64, filename = _chat_file_data_from_url(FILE_URL)
            except Exception as exc:
                raise CaseSkipped(f"C.file_input could not prepare file payload: {exc}") from exc

            messages: list[ChatCompletionMessageParam] = [
                _chat_user_message(
                    [
                        _chat_text_part("Summarize the file."),
                        _chat_file_part_from_data(file_data_b64=file_data_b64, filename=filename),
                    ]
                )
            ]
            resp = client.chat.completions.create(
                model=model,
                messages=messages,
            )
            return {"id": getattr(resp, "id", None), "content": resp.choices[0].message.content}

        def c_file_input_id() -> dict[str, Any]:
            resp = client.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "Summarize the file."},
                            {"type": "file", "file": {"file_id": CHAT_FILE_ID}},
                        ],
                    }
                ],
            )
            return {"id": getattr(resp, "id", None), "content": resp.choices[0].message.content}

        def c_tools_function() -> dict[str, Any]:
            resp = client.chat.completions.create(
                model=model,
                messages=[{"role": "user", "content": "What is the weather in Boston?"}],
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": "get_current_weather",
                            "description": "Get the current weather",
                            "parameters": {
                                "type": "object",
                                "properties": {
                                    "location": {"type": "string"},
                                    "unit": {"type": "string", "enum": ["c", "f"]},
                                },
                                "required": ["location"],
                            },
                        },
                    }
                ],
                tool_choice="auto",
            )
            return {"id": getattr(resp, "id", None), "finish_reason": resp.choices[0].finish_reason}

        def c_response_format_json_object() -> dict[str, Any]:
            resp = client.chat.completions.create(
                model=model,
                messages=[{"role": "user", "content": 'Return {"ok": true} as JSON.'}],
                response_format={"type": "json_object"},
            )
            return {"id": getattr(resp, "id", None), "content": resp.choices[0].message.content}

        def c_streaming() -> dict[str, Any]:
            stream = client.chat.completions.create(
                model=model,
                messages=[{"role": "user", "content": "Stream ok"}],
                stream=True,
            )
            return _chat_stream_summary(stream)

        cases.extend(
            [
                ("C.text", c_text),
                ("C.image_url", c_image_url),
                ("C.audio_input", c_audio_input),
                ("C.file_input", c_file_input),
                ("C.file_input_id", c_file_input_id),
                ("C.tools_function", c_tools_function),
                ("C.response_format_json_object", c_response_format_json_object),
                ("C.streaming", c_streaming),
            ]
        )

    for name, fn in cases:
        result = run_case(name, fn)
        results.append(result)
        status = result.status.upper()
        summary = result.detail.get("reason") or result.detail.get("message") or ""
        logger.info("- %s: %s%s", name, status, f" {summary}" if summary else "")

    output = {
        "timestamp": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        "base_url": BASE_URL,
        "model": model,
        "expected_unsupported": EXPECTED_UNSUPPORTED,
        "results": [asdict(item) for item in results],
    }

    os.makedirs("refs", exist_ok=True)
    with open("refs/openai-compat-live-results.json", "w", encoding="utf-8") as handle:
        json.dump(output, handle, ensure_ascii=True, indent=2)

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import builtins
import errno
import json
import logging
import sys
from types import ModuleType, SimpleNamespace
from unittest.mock import AsyncMock, Mock

import pytest

import app.core.tracing.otel as otel
from app.core.runtime_logging import JsonFormatter
from app.modules.proxy.ring_membership import RING_STALE_GRACE_SECONDS, RING_STALE_THRESHOLD_SECONDS

pytestmark = pytest.mark.unit


@pytest.fixture(autouse=True)
def reset_otel_state(monkeypatch: pytest.MonkeyPatch):
    otel._otel_initialized = False
    for name in list(sys.modules):
        if name == "opentelemetry" or name.startswith("opentelemetry."):
            monkeypatch.delitem(sys.modules, name, raising=False)
    yield
    otel._otel_initialized = False


def _install_fake_otel(monkeypatch: pytest.MonkeyPatch) -> SimpleNamespace:
    state = SimpleNamespace(
        provider=None,
        exporter_endpoint=None,
        fastapi_instrumented=0,
        aiohttp_instrumented=0,
        sqlalchemy_instrumented=0,
        resource_attributes=None,
    )

    class FakeSpanContext:
        def __init__(self, trace_id: int, span_id: int, is_valid: bool = True) -> None:
            self.trace_id = trace_id
            self.span_id = span_id
            self.is_valid = is_valid

    class FakeSpan:
        def __init__(self, context: FakeSpanContext) -> None:
            self._context = context

        def get_span_context(self) -> FakeSpanContext:
            return self._context

    class FakeTracerProvider:
        def __init__(self, *, resource: object | None = None) -> None:
            self.processors: list[FakeBatchSpanProcessor] = []
            self.resource = resource
            state.resource_attributes = getattr(resource, "attributes", None)

        def add_span_processor(self, processor: FakeBatchSpanProcessor) -> None:
            self.processors.append(processor)

    class FakeBatchSpanProcessor:
        def __init__(self, exporter: object) -> None:
            self.exporter = exporter

    class FakeOTLPSpanExporter:
        def __init__(self, endpoint: str) -> None:
            state.exporter_endpoint = endpoint
            self.endpoint = endpoint

    class FakeFastAPIInstrumentor:
        def instrument(self) -> None:
            state.fastapi_instrumented += 1

    class FakeAioHttpClientInstrumentor:
        def instrument(self) -> None:
            state.aiohttp_instrumented += 1

    class FakeSQLAlchemyInstrumentor:
        def instrument(self) -> None:
            state.sqlalchemy_instrumented += 1

    trace_module = ModuleType("opentelemetry.trace")
    setattr(trace_module, "_current_span", FakeSpan(FakeSpanContext(trace_id=0x1234, span_id=0x5678)))

    def set_tracer_provider(provider: FakeTracerProvider) -> None:
        state.provider = provider

    def get_current_span() -> FakeSpan:
        return getattr(trace_module, "_current_span")

    setattr(trace_module, "set_tracer_provider", set_tracer_provider)
    setattr(trace_module, "get_current_span", get_current_span)

    opentelemetry_module = ModuleType("opentelemetry")
    opentelemetry_module.__path__ = []
    setattr(opentelemetry_module, "trace", trace_module)

    sdk_module = ModuleType("opentelemetry.sdk")
    sdk_module.__path__ = []
    sdk_resources_module = ModuleType("opentelemetry.sdk.resources")

    class FakeResource:
        def __init__(self, attributes: dict[str, str]) -> None:
            self.attributes = attributes

        @classmethod
        def create(cls, attributes: dict[str, str]) -> "FakeResource":
            return cls(attributes)

    setattr(sdk_resources_module, "Resource", FakeResource)
    setattr(sdk_resources_module, "SERVICE_NAME", "service.name")
    sdk_trace_module = ModuleType("opentelemetry.sdk.trace")
    setattr(sdk_trace_module, "TracerProvider", FakeTracerProvider)
    sdk_trace_export_module = ModuleType("opentelemetry.sdk.trace.export")
    setattr(sdk_trace_export_module, "BatchSpanProcessor", FakeBatchSpanProcessor)

    exporter_module = ModuleType("opentelemetry.exporter")
    exporter_module.__path__ = []
    exporter_otlp_module = ModuleType("opentelemetry.exporter.otlp")
    exporter_otlp_module.__path__ = []
    exporter_otlp_proto_module = ModuleType("opentelemetry.exporter.otlp.proto")
    exporter_otlp_proto_module.__path__ = []
    exporter_otlp_proto_grpc_module = ModuleType("opentelemetry.exporter.otlp.proto.grpc")
    exporter_otlp_proto_grpc_module.__path__ = []
    exporter_trace_module = ModuleType("opentelemetry.exporter.otlp.proto.grpc.trace_exporter")
    setattr(exporter_trace_module, "OTLPSpanExporter", FakeOTLPSpanExporter)

    instrumentation_module = ModuleType("opentelemetry.instrumentation")
    instrumentation_module.__path__ = []
    instrumentation_fastapi_module = ModuleType("opentelemetry.instrumentation.fastapi")
    setattr(instrumentation_fastapi_module, "FastAPIInstrumentor", FakeFastAPIInstrumentor)
    instrumentation_aiohttp_module = ModuleType("opentelemetry.instrumentation.aiohttp_client")
    setattr(instrumentation_aiohttp_module, "AioHttpClientInstrumentor", FakeAioHttpClientInstrumentor)
    instrumentation_sqlalchemy_module = ModuleType("opentelemetry.instrumentation.sqlalchemy")
    setattr(instrumentation_sqlalchemy_module, "SQLAlchemyInstrumentor", FakeSQLAlchemyInstrumentor)

    modules = {
        "opentelemetry": opentelemetry_module,
        "opentelemetry.trace": trace_module,
        "opentelemetry.sdk": sdk_module,
        "opentelemetry.sdk.resources": sdk_resources_module,
        "opentelemetry.sdk.trace": sdk_trace_module,
        "opentelemetry.sdk.trace.export": sdk_trace_export_module,
        "opentelemetry.exporter": exporter_module,
        "opentelemetry.exporter.otlp": exporter_otlp_module,
        "opentelemetry.exporter.otlp.proto": exporter_otlp_proto_module,
        "opentelemetry.exporter.otlp.proto.grpc": exporter_otlp_proto_grpc_module,
        "opentelemetry.exporter.otlp.proto.grpc.trace_exporter": exporter_trace_module,
        "opentelemetry.instrumentation": instrumentation_module,
        "opentelemetry.instrumentation.fastapi": instrumentation_fastapi_module,
        "opentelemetry.instrumentation.aiohttp_client": instrumentation_aiohttp_module,
        "opentelemetry.instrumentation.sqlalchemy": instrumentation_sqlalchemy_module,
    }

    for name, module in modules.items():
        monkeypatch.setitem(sys.modules, name, module)

    return state


def test_init_tracing_returns_false_when_opentelemetry_is_unavailable(monkeypatch: pytest.MonkeyPatch):
    original_import = builtins.__import__

    def raising_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name == "opentelemetry" or name.startswith("opentelemetry."):
            raise ImportError("missing opentelemetry")
        return original_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, "__import__", raising_import)

    assert otel.init_tracing() is False
    assert otel.is_initialized() is False


def test_init_tracing_returns_true_when_opentelemetry_modules_are_available(monkeypatch: pytest.MonkeyPatch):
    state = _install_fake_otel(monkeypatch)

    assert otel.init_tracing(service_name="codex-lb", endpoint="http://collector:4317") is True
    assert otel.is_initialized() is True
    assert state.provider is not None
    assert len(state.provider.processors) == 1
    assert state.exporter_endpoint == "http://collector:4317"
    assert state.resource_attributes == {"service.name": "codex-lb"}
    assert state.fastapi_instrumented == 1
    assert state.aiohttp_instrumented == 1
    assert state.sqlalchemy_instrumented == 1


def test_get_current_trace_id_returns_none_when_opentelemetry_is_inactive(monkeypatch: pytest.MonkeyPatch):
    original_import = builtins.__import__

    def raising_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name == "opentelemetry" or name.startswith("opentelemetry."):
            raise ImportError("missing opentelemetry")
        return original_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, "__import__", raising_import)

    assert otel.get_current_trace_id() is None
    assert otel.get_current_span_id() is None


def test_json_formatter_includes_trace_and_span_ids_when_available(monkeypatch: pytest.MonkeyPatch):
    _install_fake_otel(monkeypatch)

    record = logging.LogRecord(
        name="test.module",
        level=logging.INFO,
        pathname=__file__,
        lineno=1,
        msg="hello",
        args=(),
        exc_info=None,
    )

    parsed = json.loads(JsonFormatter().format(record))

    assert parsed["trace_id"] == "00000000000000000000000000001234"
    assert parsed["span_id"] == "0000000000005678"


class _DummyScheduler:
    def __init__(self) -> None:
        self.started = False
        self.stopped = False

    async def start(self) -> None:
        self.started = True

    async def stop(self) -> None:
        self.stopped = True


@pytest.mark.asyncio
async def test_lifespan_runs_normally_when_otel_is_disabled(monkeypatch: pytest.MonkeyPatch):
    import app.core.startup as startup_module
    import app.main as main

    settings = SimpleNamespace(
        otel_enabled=False,
        otel_exporter_endpoint="",
        metrics_enabled=False,
        shutdown_drain_timeout_seconds=0,
    )
    settings_cache = SimpleNamespace(invalidate=AsyncMock())
    rate_limit_cache = SimpleNamespace(invalidate=AsyncMock())
    usage_scheduler = _DummyScheduler()
    model_scheduler = _DummyScheduler()
    sticky_scheduler = _DummyScheduler()
    call_order: list[str] = []

    async def _init_db() -> None:
        call_order.append("init_db")

    def _init_background_db() -> None:
        call_order.append("init_background_db")

    init_db = AsyncMock()
    init_db.side_effect = _init_db
    init_background_db = Mock(side_effect=_init_background_db)
    init_http_client = AsyncMock()
    close_http_client = AsyncMock()
    close_db = AsyncMock()

    monkeypatch.setattr(main, "get_settings", lambda: settings)
    monkeypatch.setattr(main, "get_settings_cache", lambda: settings_cache)
    monkeypatch.setattr(main, "get_rate_limit_headers_cache", lambda: rate_limit_cache)
    monkeypatch.setattr(main, "reload_additional_quota_registry", lambda: None)
    monkeypatch.setattr(main, "init_db", init_db)
    monkeypatch.setattr(main, "init_background_db", init_background_db)
    monkeypatch.setattr(main, "init_http_client", init_http_client)
    monkeypatch.setattr(main, "close_http_client", close_http_client)
    monkeypatch.setattr(main, "close_db", close_db)
    monkeypatch.setattr(main, "build_usage_refresh_scheduler", lambda: usage_scheduler)
    monkeypatch.setattr(main, "build_model_refresh_scheduler", lambda: model_scheduler)
    monkeypatch.setattr(main, "build_sticky_session_cleanup_scheduler", lambda: sticky_scheduler)

    async with main.lifespan(main.app):
        assert startup_module._startup_complete is True
        assert usage_scheduler.started is True
        assert model_scheduler.started is True
        assert sticky_scheduler.started is True

    init_db.assert_awaited_once()
    init_background_db.assert_called_once()
    init_http_client.assert_awaited_once()
    close_http_client.assert_awaited_once()
    close_db.assert_awaited_once()
    settings_cache.invalidate.assert_awaited_once()
    rate_limit_cache.invalidate.assert_awaited_once()
    assert call_order[:2] == ["init_db", "init_background_db"]
    assert usage_scheduler.stopped is True
    assert model_scheduler.stopped is True
    assert sticky_scheduler.stopped is True


@pytest.mark.asyncio
async def test_lifespan_marks_bridge_membership_stale_on_shutdown(monkeypatch: pytest.MonkeyPatch):
    import app.main as main

    settings = SimpleNamespace(
        otel_enabled=False,
        otel_exporter_endpoint="",
        metrics_enabled=False,
        shutdown_drain_timeout_seconds=0,
        http_responses_session_bridge_instance_id="pod-a",
    )
    settings_cache = SimpleNamespace(invalidate=AsyncMock())
    rate_limit_cache = SimpleNamespace(invalidate=AsyncMock())
    usage_scheduler = _DummyScheduler()
    model_scheduler = _DummyScheduler()
    sticky_scheduler = _DummyScheduler()
    close_http_client = AsyncMock()
    close_db = AsyncMock()
    ring_service = SimpleNamespace(
        register=AsyncMock(),
        mark_stale=AsyncMock(),
        unregister=AsyncMock(),
        heartbeat=AsyncMock(),
    )
    cache_poller = SimpleNamespace(
        on_invalidation=Mock(),
        start=AsyncMock(),
        stop=AsyncMock(),
    )

    monkeypatch.setattr(main, "get_settings", lambda: settings)
    monkeypatch.setattr(main, "get_settings_cache", lambda: settings_cache)
    monkeypatch.setattr(main, "get_rate_limit_headers_cache", lambda: rate_limit_cache)
    monkeypatch.setattr(main, "reload_additional_quota_registry", lambda: None)
    monkeypatch.setattr(main, "init_db", AsyncMock())
    monkeypatch.setattr(main, "init_background_db", Mock())
    monkeypatch.setattr(main, "init_http_client", AsyncMock())
    monkeypatch.setattr(main, "close_http_client", close_http_client)
    monkeypatch.setattr(main, "close_db", close_db)
    monkeypatch.setattr(main, "build_usage_refresh_scheduler", lambda: usage_scheduler)
    monkeypatch.setattr(main, "build_model_refresh_scheduler", lambda: model_scheduler)
    monkeypatch.setattr(main, "build_sticky_session_cleanup_scheduler", lambda: sticky_scheduler)
    monkeypatch.setattr(main, "RingMembershipService", lambda session_factory: ring_service)
    monkeypatch.setattr(main, "mark_process_dead", Mock())
    monkeypatch.setattr(
        "app.core.cache.invalidation.CacheInvalidationPoller",
        lambda session_factory: cache_poller,
    )

    async with main.lifespan(main.app):
        pass

    ring_service.register.assert_awaited_once_with("pod-a")
    ring_service.mark_stale.assert_awaited_once_with(
        "pod-a",
        stale_threshold_seconds=RING_STALE_THRESHOLD_SECONDS,
        grace_seconds=RING_STALE_GRACE_SECONDS,
    )
    ring_service.unregister.assert_not_called()


@pytest.mark.asyncio
async def test_lifespan_skips_dynamic_bridge_ring_membership_when_disabled(monkeypatch: pytest.MonkeyPatch):
    import app.main as main

    settings = SimpleNamespace(
        otel_enabled=False,
        otel_exporter_endpoint="",
        metrics_enabled=False,
        shutdown_drain_timeout_seconds=0,
        http_responses_session_bridge_instance_id="pod-a",
        http_responses_session_bridge_enabled=True,
        database_url="sqlite+aiosqlite:///./store.db",
        http_responses_session_bridge_instance_ring=[],
    )
    settings_cache = SimpleNamespace(invalidate=AsyncMock())
    rate_limit_cache = SimpleNamespace(invalidate=AsyncMock())
    usage_scheduler = _DummyScheduler()
    model_scheduler = _DummyScheduler()
    sticky_scheduler = _DummyScheduler()
    close_http_client = AsyncMock()
    close_db = AsyncMock()
    cache_poller = SimpleNamespace(
        on_invalidation=Mock(),
        start=AsyncMock(),
        stop=AsyncMock(),
    )
    ring_service_factory = Mock(side_effect=AssertionError("dynamic ring membership should stay disabled"))

    monkeypatch.setattr(main, "get_settings", lambda: settings)
    monkeypatch.setattr(main, "get_settings_cache", lambda: settings_cache)
    monkeypatch.setattr(main, "get_rate_limit_headers_cache", lambda: rate_limit_cache)
    monkeypatch.setattr(main, "reload_additional_quota_registry", lambda: None)
    monkeypatch.setattr(main, "init_db", AsyncMock())
    monkeypatch.setattr(main, "init_background_db", Mock())
    monkeypatch.setattr(main, "init_http_client", AsyncMock())
    monkeypatch.setattr(main, "close_http_client", close_http_client)
    monkeypatch.setattr(main, "close_db", close_db)
    monkeypatch.setattr(main, "build_usage_refresh_scheduler", lambda: usage_scheduler)
    monkeypatch.setattr(main, "build_model_refresh_scheduler", lambda: model_scheduler)
    monkeypatch.setattr(main, "build_sticky_session_cleanup_scheduler", lambda: sticky_scheduler)
    monkeypatch.setattr(main, "dynamic_bridge_ring_membership_enabled", lambda current: False)
    monkeypatch.setattr(main, "RingMembershipService", ring_service_factory)
    monkeypatch.setattr(main, "mark_process_dead", Mock())
    monkeypatch.setattr(
        "app.core.cache.invalidation.CacheInvalidationPoller",
        lambda session_factory: cache_poller,
    )

    async with main.lifespan(main.app):
        pass

    ring_service_factory.assert_not_called()


def test_metrics_bind_failure_is_only_benign_in_multiprocess_mode(monkeypatch: pytest.MonkeyPatch):
    import app.main as main

    monkeypatch.setattr(main, "MULTIPROCESS_MODE", False)
    assert main._is_benign_metrics_bind_failure(SystemExit(1)) is False
    assert main._is_benign_metrics_bind_failure(OSError(errno.EADDRINUSE, "in use")) is False

    monkeypatch.setattr(main, "MULTIPROCESS_MODE", True)
    assert main._is_benign_metrics_bind_failure(SystemExit(1)) is True
    assert main._is_benign_metrics_bind_failure(OSError(errno.EADDRINUSE, "in use")) is True



from __future__ import annotations


class AppError(Exception):
    """Base exception for all domain errors."""

    status_code: int = 500
    code: str = "internal_error"
    message: str = "Unexpected error"

    def __init__(self, message: str | None = None, *, code: str | None = None) -> None:
        self.message = message or self.__class__.message
        if code is not None:
            self.code = code
        super().__init__(self.message)


# --- OpenAI-envelope errors (proxy routes) ---


class ProxyAuthError(AppError):
    status_code = 401
    code = "invalid_api_key"
    error_type = "authentication_error"


class ProxyModelNotAllowed(AppError):
    status_code = 403
    code = "model_not_allowed"
    error_type = "permission_error"


class ProxyRateLimitError(AppError):
    status_code = 429
    code = "rate_limit_exceeded"
    error_type = "rate_limit_error"


class ProxyUpstreamError(AppError):
    status_code = 503
    code = "upstream_error"
    error_type = "server_error"


# --- Dashboard-envelope errors ---


class DashboardAuthError(AppError):
    status_code = 401
    code = "authentication_required"


class DashboardNotFoundError(AppError):
    status_code = 404
    code = "not_found"


class DashboardConflictError(AppError):
    status_code = 409
    code = "conflict"


class DashboardBadRequestError(AppError):
    status_code = 400
    code = "bad_request"


class DashboardValidationError(AppError):
    status_code = 422
    code = "validation_error"


class DashboardRateLimitError(AppError):
    status_code = 429
    code = "rate_limited"

    def __init__(self, message: str, *, retry_after: int, code: str | None = None) -> None:
        self.retry_after = retry_after
        super().__init__(message, code=code)



from __future__ import annotations

from collections.abc import Mapping

type JsonValue = bool | int | float | str | None | list[JsonValue] | Mapping[str, JsonValue]
type JsonObject = Mapping[str, JsonValue]



from __future__ import annotations

import base64
import json
from datetime import datetime, timedelta, timezone

import pytest
from sqlalchemy import update

from app.core.auth import fallback_account_id, generate_unique_account_id
from app.core.crypto import TokenEncryptor
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus
from app.db.session import SessionLocal
from app.modules.accounts.repository import AccountsRepository
from app.modules.request_logs.repository import RequestLogsRepository
from app.modules.usage.repository import UsageRepository

pytestmark = pytest.mark.integration


def _encode_jwt(payload: dict) -> str:
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    body = base64.urlsafe_b64encode(raw).rstrip(b"=").decode("ascii")
    return f"header.{body}.sig"


def _make_auth_json(account_id: str | None, email: str, plan_type: str = "plus") -> dict:
    payload = {
        "email": email,
        "https://api.openai.com/auth": {"chatgpt_plan_type": plan_type},
    }
    if account_id:
        payload["chatgpt_account_id"] = account_id
    tokens: dict[str, object] = {
        "idToken": _encode_jwt(payload),
        "accessToken": "access",
        "refreshToken": "refresh",
    }
    if account_id:
        tokens["accountId"] = account_id
    return {"tokens": tokens}


def _make_account(account_id: str, email: str, plan_type: str = "plus") -> Account:
    encryptor = TokenEncryptor()
    return Account(
        id=account_id,
        email=email,
        plan_type=plan_type,
        access_token_encrypted=encryptor.encrypt("access"),
        refresh_token_encrypted=encryptor.encrypt("refresh"),
        id_token_encrypted=encryptor.encrypt("id"),
        last_refresh=utcnow(),
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )


def _iso_utc(epoch_seconds: int) -> str:
    return (
        datetime.fromtimestamp(epoch_seconds, tz=timezone.utc)
        .isoformat()
        .replace(
            "+00:00",
            "Z",
        )
    )


@pytest.mark.asyncio
async def test_import_invalid_json_returns_400(async_client):
    files = {"auth_json": ("auth.json", "not-json", "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "invalid_auth_json"


@pytest.mark.asyncio
async def test_import_missing_tokens_returns_400(async_client):
    files = {"auth_json": ("auth.json", json.dumps({"foo": "bar"}), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 400
    payload = response.json()
    assert payload["error"]["code"] == "invalid_auth_json"


@pytest.mark.asyncio
async def test_import_falls_back_to_email_based_account_id(async_client):
    email = "fallback@example.com"
    auth_json = _make_auth_json(None, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200
    payload = response.json()
    assert payload["accountId"] == fallback_account_id(email)
    assert payload["email"] == email


@pytest.mark.asyncio
async def test_import_overwrites_for_same_account_identity_when_overwrite_enabled(async_client):
    settings = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "importWithoutOverwrite": False,
            "totpRequiredOnLogin": False,
        },
    )
    assert settings.status_code == 200
    assert settings.json()["importWithoutOverwrite"] is False

    email = "same-default@example.com"
    raw_account_id = "acc_same_default"

    files_one = {
        "auth_json": (
            "auth.json",
            json.dumps(_make_auth_json(raw_account_id, email, "plus")),
            "application/json",
        )
    }
    first = await async_client.post("/api/accounts/import", files=files_one)
    assert first.status_code == 200

    files_two = {
        "auth_json": (
            "auth.json",
            json.dumps(_make_auth_json(raw_account_id, email, "team")),
            "application/json",
        )
    }
    second = await async_client.post("/api/accounts/import", files=files_two)
    assert second.status_code == 200

    expected_account_id = generate_unique_account_id(raw_account_id, email)
    assert first.json()["accountId"] == expected_account_id
    assert second.json()["accountId"] == expected_account_id
    assert second.json()["planType"] == "team"

    accounts_response = await async_client.get("/api/accounts")
    assert accounts_response.status_code == 200
    accounts = [entry for entry in accounts_response.json()["accounts"] if entry["email"] == email]
    assert len(accounts) == 1
    assert accounts[0]["accountId"] == expected_account_id
    assert accounts[0]["planType"] == "team"


@pytest.mark.asyncio
async def test_import_without_overwrite_keeps_same_account_identity_separate(async_client):
    settings = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "importWithoutOverwrite": True,
            "totpRequiredOnLogin": False,
        },
    )
    assert settings.status_code == 200
    assert settings.json()["importWithoutOverwrite"] is True

    email = "same-separate@example.com"
    raw_account_id = "acc_same_separate"

    files_one = {
        "auth_json": (
            "auth.json",
            json.dumps(_make_auth_json(raw_account_id, email, "plus")),
            "application/json",
        )
    }
    first = await async_client.post("/api/accounts/import", files=files_one)
    assert first.status_code == 200

    files_two = {
        "auth_json": (
            "auth.json",
            json.dumps(_make_auth_json(raw_account_id, email, "team")),
            "application/json",
        )
    }
    second = await async_client.post("/api/accounts/import", files=files_two)
    assert second.status_code == 200

    base_account_id = generate_unique_account_id(raw_account_id, email)
    first_id = first.json()["accountId"]
    second_id = second.json()["accountId"]
    assert first_id == base_account_id
    assert second_id != first_id
    assert second_id.startswith(f"{base_account_id}__copy")

    accounts_response = await async_client.get("/api/accounts")
    assert accounts_response.status_code == 200
    accounts = [entry for entry in accounts_response.json()["accounts"] if entry["email"] == email]
    assert len(accounts) == 2
    ids = {entry["accountId"] for entry in accounts}
    assert ids == {first_id, second_id}


@pytest.mark.asyncio
async def test_import_returns_409_when_overwrite_mode_cannot_resolve_duplicate_email(async_client):
    enable_separate = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "importWithoutOverwrite": True,
            "totpRequiredOnLogin": False,
        },
    )
    assert enable_separate.status_code == 200
    assert enable_separate.json()["importWithoutOverwrite"] is True

    email = "conflict@example.com"
    raw_account_id = "acc_conflict_base"

    first = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json(raw_account_id, email, "plus")),
                "application/json",
            )
        },
    )
    assert first.status_code == 200

    second = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json(raw_account_id, email, "team")),
                "application/json",
            )
        },
    )
    assert second.status_code == 200
    assert second.json()["accountId"] != first.json()["accountId"]

    enable_overwrite = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "importWithoutOverwrite": False,
            "totpRequiredOnLogin": False,
        },
    )
    assert enable_overwrite.status_code == 200
    assert enable_overwrite.json()["importWithoutOverwrite"] is False

    conflict = await async_client.post(
        "/api/accounts/import",
        files={
            "auth_json": (
                "auth.json",
                json.dumps(_make_auth_json("acc_conflict_new", email, "pro")),
                "application/json",
            )
        },
    )
    assert conflict.status_code == 409
    payload = conflict.json()
    assert payload["error"]["code"] == "duplicate_identity_conflict"

    accounts_response = await async_client.get("/api/accounts")
    assert accounts_response.status_code == 200
    accounts = [entry for entry in accounts_response.json()["accounts"] if entry["email"] == email]
    assert len(accounts) == 2
    assert all(entry["planType"] != "pro" for entry in accounts)


@pytest.mark.asyncio
async def test_delete_account_removes_from_list(async_client):
    email = "delete@example.com"
    raw_account_id = "acc_delete"
    auth_json = _make_auth_json(raw_account_id, email)
    files = {"auth_json": ("auth.json", json.dumps(auth_json), "application/json")}
    response = await async_client.post("/api/accounts/import", files=files)
    assert response.status_code == 200

    actual_account_id = generate_unique_account_id(raw_account_id, email)
    delete = await async_client.delete(f"/api/accounts/{actual_account_id}")
    assert delete.status_code == 200
    assert delete.json()["status"] == "deleted"

    accounts = await async_client.get("/api/accounts")
    assert accounts.status_code == 200
    data = accounts.json()["accounts"]
    assert all(account["accountId"] != actual_account_id for account in data)


@pytest.mark.asyncio
async def test_accounts_list_includes_per_account_reset_times(async_client, db_setup):
    primary_a = 1735689600
    primary_b = 1735693200
    secondary_a = 1736294400
    secondary_b = 1736380800

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_reset_a", "a@example.com"))
        await accounts_repo.upsert(_make_account("acc_reset_b", "b@example.com"))

        await usage_repo.add_entry(
            "acc_reset_a",
            10.0,
            window="primary",
            reset_at=primary_a,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            "acc_reset_b",
            20.0,
            window="primary",
            reset_at=primary_b,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            "acc_reset_a",
            30.0,
            window="secondary",
            reset_at=secondary_a,
            window_minutes=10080,
        )
        await usage_repo.add_entry(
            "acc_reset_b",
            40.0,
            window="secondary",
            reset_at=secondary_b,
            window_minutes=10080,
        )

    response = await async_client.get("/api/accounts")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}

    assert accounts["acc_reset_a"]["resetAtPrimary"] == _iso_utc(primary_a)
    assert accounts["acc_reset_b"]["resetAtPrimary"] == _iso_utc(primary_b)
    assert accounts["acc_reset_a"]["resetAtSecondary"] == _iso_utc(secondary_a)
    assert accounts["acc_reset_b"]["resetAtSecondary"] == _iso_utc(secondary_b)
    assert accounts["acc_reset_a"]["windowMinutesPrimary"] == 300
    assert accounts["acc_reset_b"]["windowMinutesPrimary"] == 300
    assert accounts["acc_reset_a"]["windowMinutesSecondary"] == 10080
    assert accounts["acc_reset_b"]["windowMinutesSecondary"] == 10080


@pytest.mark.asyncio
async def test_accounts_list_includes_request_usage_cost_rollup(async_client, db_setup):
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)

        await accounts_repo.upsert(_make_account("acc_cost", "cost@example.com"))
        await accounts_repo.upsert(_make_account("acc_other", "other@example.com"))

        await logs_repo.add_log(
            account_id="acc_cost",
            request_id="req_cost_1",
            model="gpt-5.3-codex",
            input_tokens=100_000,
            output_tokens=20_000,
            cached_input_tokens=90_000,
            latency_ms=200,
            status="success",
            error_code=None,
        )
        await logs_repo.add_log(
            account_id="acc_cost",
            request_id="req_cost_2",
            model="gpt-5.1-codex",
            input_tokens=50_000,
            output_tokens=10_000,
            cached_input_tokens=0,
            latency_ms=180,
            status="success",
            error_code=None,
        )
        await logs_repo.add_log(
            account_id="acc_other",
            request_id="req_other_1",
            model="gpt-5.1-codex-mini",
            input_tokens=1_000,
            output_tokens=500,
            cached_input_tokens=0,
            latency_ms=150,
            status="success",
            error_code=None,
        )

    response = await async_client.get("/api/accounts")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}

    request_usage = accounts["acc_cost"]["requestUsage"]
    assert request_usage is not None
    assert request_usage["requestCount"] == 2
    assert request_usage["totalTokens"] == 180_000
    assert request_usage["cachedInputTokens"] == 90_000
    assert request_usage["totalCostUsd"] == pytest.approx(0.47575, abs=1e-6)

    other_usage = accounts["acc_other"]["requestUsage"]
    assert other_usage is not None
    assert other_usage["requestCount"] == 1
    assert other_usage["totalTokens"] == 1_500


@pytest.mark.asyncio
async def test_accounts_list_request_usage_cost_rollup_respects_service_tier(async_client, db_setup):
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)

        await accounts_repo.upsert(_make_account("acc_priority_cost", "priority-cost@example.com"))

        await logs_repo.add_log(
            account_id="acc_priority_cost",
            request_id="req_priority_cost_1",
            model="gpt-5.4",
            service_tier="priority",
            input_tokens=1_000_000,
            output_tokens=1_000_000,
            latency_ms=200,
            status="success",
            error_code=None,
        )

    response = await async_client.get("/api/accounts")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}

    request_usage = accounts["acc_priority_cost"]["requestUsage"]
    assert request_usage is not None
    assert request_usage["requestCount"] == 1
    assert request_usage["totalTokens"] == 2_000_000
    assert request_usage["cachedInputTokens"] == 0
    assert request_usage["totalCostUsd"] == pytest.approx(35.0, abs=1e-6)


@pytest.mark.asyncio
async def test_accounts_list_request_usage_uses_persisted_cost(async_client, db_setup):
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        logs_repo = RequestLogsRepository(session)

        await accounts_repo.upsert(_make_account("acc_persisted_cost", "persisted-cost@example.com"))

        log = await logs_repo.add_log(
            account_id="acc_persisted_cost",
            request_id="req_persisted_cost_1",
            model="gpt-5.1",
            input_tokens=10,
            output_tokens=5,
            latency_ms=50,
            status="success",
            error_code=None,
        )
        await session.execute(update(log.__class__).where(log.__class__.id == log.id).values(cost_usd=12.345678))
        await session.commit()

    response = await async_client.get("/api/accounts")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}

    request_usage = accounts["acc_persisted_cost"]["requestUsage"]
    assert request_usage is not None
    assert request_usage["totalCostUsd"] == pytest.approx(12.345678, abs=1e-6)


@pytest.mark.asyncio
async def test_accounts_list_maps_weekly_only_primary_to_secondary(async_client, db_setup):
    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_free_like", "free@example.com", plan_type="free"))
        await usage_repo.add_entry(
            "acc_free_like",
            24.0,
            window="primary",
            window_minutes=10080,
        )

    response = await async_client.get("/api/accounts")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}

    account = accounts["acc_free_like"]
    assert account["usage"]["primaryRemainingPercent"] is None
    assert account["usage"]["secondaryRemainingPercent"] == pytest.approx(76.0)
    assert account["windowMinutesPrimary"] is None
    assert account["windowMinutesSecondary"] == 10080


@pytest.mark.asyncio
async def test_accounts_list_prefers_newer_weekly_primary_over_stale_secondary(async_client, db_setup):
    now = utcnow()
    stale_reset = 1735689600
    weekly_reset = 1735862400

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)

        await accounts_repo.upsert(_make_account("acc_weekly_stale", "weekly-stale@example.com", plan_type="free"))
        await usage_repo.add_entry(
            "acc_weekly_stale",
            15.0,
            window="secondary",
            reset_at=stale_reset,
            window_minutes=10080,
            recorded_at=now - timedelta(days=2),
        )
        await usage_repo.add_entry(
            "acc_weekly_stale",
            80.0,
            window="primary",
            reset_at=weekly_reset,
            window_minutes=10080,
            recorded_at=now,
        )

    response = await async_client.get("/api/accounts")
    assert response.status_code == 200
    payload = response.json()
    accounts = {item["accountId"]: item for item in payload["accounts"]}

    account = accounts["acc_weekly_stale"]
    assert account["usage"]["primaryRemainingPercent"] is None
    assert account["usage"]["secondaryRemainingPercent"] == pytest.approx(20.0)
    assert account["windowMinutesPrimary"] is None
    assert account["windowMinutesSecondary"] == 10080
    assert account["resetAtSecondary"] == _iso_utc(weekly_reset)



from __future__ import annotations

from datetime import UTC, datetime, timedelta

from sqlalchemy import delete, func, insert, literal, select, text
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.exceptions import DashboardRateLimitError
from app.db.models import RateLimitAttempt


class DatabaseRateLimiter:
    def __init__(self, max_attempts: int, window_seconds: int, type: str) -> None:
        if max_attempts <= 0:
            raise ValueError("max_attempts must be positive")
        if window_seconds <= 0:
            raise ValueError("window_seconds must be positive")
        self.max_attempts = max_attempts
        self.window_seconds = window_seconds
        self.type = type

    def _window_start(self) -> datetime:
        return datetime.now(UTC) - timedelta(seconds=self.window_seconds)

    async def check(self, key: str, session: AsyncSession) -> None:
        """Read-only rate-limit check. Raises if limit is exceeded; does NOT record an attempt."""
        window_start = self._window_start()

        count = await session.scalar(
            select(func.count())
            .select_from(RateLimitAttempt)
            .where(
                RateLimitAttempt.key == key,
                RateLimitAttempt.type == self.type,
                RateLimitAttempt.attempted_at >= window_start,
            )
        )

        if (count or 0) >= self.max_attempts:
            oldest_attempt = await session.scalar(
                select(RateLimitAttempt.attempted_at)
                .where(
                    RateLimitAttempt.key == key,
                    RateLimitAttempt.type == self.type,
                    RateLimitAttempt.attempted_at >= window_start,
                )
                .order_by(RateLimitAttempt.attempted_at.asc())
                .limit(1)
            )
            now = datetime.now(UTC)
            retry_after = self.window_seconds
            if oldest_attempt is not None:
                if oldest_attempt.tzinfo is None:
                    oldest_attempt = oldest_attempt.replace(tzinfo=UTC)
                reset_at = oldest_attempt + timedelta(seconds=self.window_seconds)
                retry_after = max(1, int((reset_at - now).total_seconds()))
            raise DashboardRateLimitError("Too many attempts", retry_after=retry_after)

    async def check_and_increment(self, key: str, session: AsyncSession) -> None:
        """Atomically check the rate limit and record an attempt.

        PostgreSQL: advisory lock serialises the count-then-insert.
        SQLite: a single INSERT … SELECT statement with a WHERE-count
        guard is processed atomically by SQLite's single-writer lock,
        safe across processes sharing the same database file.
        """
        dialect_name = session.get_bind().dialect.name
        if dialect_name == "postgresql":
            lock_key = f"{self.type}:{key}"
            await session.execute(text("SELECT pg_advisory_xact_lock(hashtext(:key))"), {"key": lock_key})

        now = datetime.now(UTC)
        window_start = now - timedelta(seconds=self.window_seconds)
        key_column = RateLimitAttempt.__table__.c.key
        type_column = RateLimitAttempt.__table__.c.type
        attempted_at_column = RateLimitAttempt.__table__.c.attempted_at

        raw_result = await session.execute(
            insert(RateLimitAttempt).from_select(
                [key_column.name, type_column.name, attempted_at_column.name],
                select(
                    literal(key, type_=key_column.type),
                    literal(self.type, type_=type_column.type),
                    literal(now, type_=attempted_at_column.type),
                ).where(
                    select(func.count())
                    .select_from(RateLimitAttempt)
                    .where(
                        RateLimitAttempt.key == key,
                        RateLimitAttempt.type == self.type,
                        RateLimitAttempt.attempted_at >= window_start,
                    )
                    .scalar_subquery()
                    < self.max_attempts
                ),
            )
        )
        inserted = raw_result.rowcount > 0  # type: ignore[union-attr]
        await session.commit()

        if not inserted:
            oldest_attempt = await session.scalar(
                select(RateLimitAttempt.attempted_at)
                .where(
                    RateLimitAttempt.key == key,
                    RateLimitAttempt.type == self.type,
                    RateLimitAttempt.attempted_at >= window_start,
                )
                .order_by(RateLimitAttempt.attempted_at.asc())
                .limit(1)
            )
            retry_after = self.window_seconds
            if oldest_attempt is not None:
                if oldest_attempt.tzinfo is None:
                    oldest_attempt = oldest_attempt.replace(tzinfo=UTC)
                reset_at = oldest_attempt + timedelta(seconds=self.window_seconds)
                retry_after = max(1, int((reset_at - now).total_seconds()))
            raise DashboardRateLimitError("Too many attempts", retry_after=retry_after)

    async def record_failure(self, key: str, session: AsyncSession) -> None:
        """Record a failed attempt atomically.

        On PostgreSQL uses an advisory lock to serialize concurrent
        record_failure calls for the same rate key.  After inserting,
        re-counts attempts; if the post-insert total exceeds
        ``max_attempts`` raises ``DashboardRateLimitError``.
        """
        now = datetime.now(UTC)
        lock_key = f"{self.type}:{key}"

        dialect_name = session.get_bind().dialect.name
        if dialect_name == "postgresql":
            await session.execute(text("SELECT pg_advisory_xact_lock(hashtext(:key))"), {"key": lock_key})

        session.add(RateLimitAttempt(key=key, type=self.type, attempted_at=now))
        await session.flush()

        window_start = now - timedelta(seconds=self.window_seconds)
        count = await session.scalar(
            select(func.count())
            .select_from(RateLimitAttempt)
            .where(
                RateLimitAttempt.key == key,
                RateLimitAttempt.type == self.type,
                RateLimitAttempt.attempted_at >= window_start,
            )
        )
        await session.commit()

        if (count or 0) > self.max_attempts:
            oldest_attempt = await session.scalar(
                select(RateLimitAttempt.attempted_at)
                .where(
                    RateLimitAttempt.key == key,
                    RateLimitAttempt.type == self.type,
                    RateLimitAttempt.attempted_at >= window_start,
                )
                .order_by(RateLimitAttempt.attempted_at.asc())
                .limit(1)
            )
            retry_after = self.window_seconds
            if oldest_attempt is not None:
                if oldest_attempt.tzinfo is None:
                    oldest_attempt = oldest_attempt.replace(tzinfo=UTC)
                reset_at = oldest_attempt + timedelta(seconds=self.window_seconds)
                retry_after = max(1, int((reset_at - now).total_seconds()))
            raise DashboardRateLimitError("Too many attempts", retry_after=retry_after)

    async def clear_for_key(self, key: str, session: AsyncSession) -> None:
        """Delete all attempts for (key, type) — used to reset counter on successful auth."""
        await session.execute(
            delete(RateLimitAttempt).where(
                RateLimitAttempt.key == key,
                RateLimitAttempt.type == self.type,
            )
        )
        await session.commit()

    async def check_and_record(self, key: str, session: AsyncSession) -> None:
        """Legacy method: check + unconditionally record. Kept for TOTP backward compatibility."""
        now = datetime.now(UTC)
        window_start = now - timedelta(seconds=self.window_seconds)

        count = await session.scalar(
            select(func.count())
            .select_from(RateLimitAttempt)
            .where(
                RateLimitAttempt.key == key,
                RateLimitAttempt.type == self.type,
                RateLimitAttempt.attempted_at >= window_start,
            )
        )

        oldest_attempt = await session.scalar(
            select(RateLimitAttempt.attempted_at)
            .where(
                RateLimitAttempt.key == key,
                RateLimitAttempt.type == self.type,
                RateLimitAttempt.attempted_at >= window_start,
            )
            .order_by(RateLimitAttempt.attempted_at.asc())
            .limit(1)
        )

        session.add(RateLimitAttempt(key=key, type=self.type, attempted_at=now))
        await session.flush()
        await session.commit()

        if (count or 0) >= self.max_attempts:
            retry_after = self.window_seconds
            if oldest_attempt is not None:
                if oldest_attempt.tzinfo is None:
                    oldest_attempt = oldest_attempt.replace(tzinfo=UTC)
                reset_at = oldest_attempt + timedelta(seconds=self.window_seconds)
                retry_after = max(1, int((reset_at - now).total_seconds()))
            raise DashboardRateLimitError("Too many attempts", retry_after=retry_after)

    async def cleanup(self, session: AsyncSession, older_than_seconds: int = 3600) -> None:
        cutoff = datetime.now(UTC) - timedelta(seconds=older_than_seconds)
        await session.execute(delete(RateLimitAttempt).where(RateLimitAttempt.attempted_at < cutoff))
        await session.commit()



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from datetime import timedelta, timezone

import pytest

from app.core.crypto import TokenEncryptor
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus
from app.db.session import SessionLocal
from app.modules.accounts.repository import AccountsRepository
from app.modules.api_keys.repository import ApiKeysRepository
from app.modules.proxy.load_balancer import LoadBalancer
from app.modules.proxy.repo_bundle import ProxyRepositories
from app.modules.proxy.sticky_repository import StickySessionsRepository
from app.modules.request_logs.repository import RequestLogsRepository
from app.modules.usage.repository import AdditionalUsageRepository, UsageRepository

pytestmark = pytest.mark.integration


@asynccontextmanager
async def _repo_factory() -> AsyncIterator[ProxyRepositories]:
    async with SessionLocal() as session:
        yield ProxyRepositories(
            accounts=AccountsRepository(session),
            usage=UsageRepository(session),
            request_logs=RequestLogsRepository(session),
            sticky_sessions=StickySessionsRepository(session),
            api_keys=ApiKeysRepository(session),
            additional_usage=AdditionalUsageRepository(session),
        )


@pytest.mark.asyncio
async def test_load_balancer_skips_secondary_quota(db_setup):
    encryptor = TokenEncryptor()
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_reset = now_epoch + 3600
    secondary_reset = now_epoch + 7200

    account_a = Account(
        id="acc_secondary_full",
        email="secondary_full@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access-a"),
        refresh_token_encrypted=encryptor.encrypt("refresh-a"),
        id_token_encrypted=encryptor.encrypt("id-a"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )
    account_b = Account(
        id="acc_secondary_ok",
        email="secondary_ok@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access-b"),
        refresh_token_encrypted=encryptor.encrypt("refresh-b"),
        id_token_encrypted=encryptor.encrypt("id-b"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)
        await accounts_repo.upsert(account_a)
        await accounts_repo.upsert(account_b)

        await usage_repo.add_entry(
            account_id=account_a.id,
            used_percent=10.0,
            window="primary",
            reset_at=primary_reset,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=account_a.id,
            used_percent=100.0,
            window="secondary",
            reset_at=secondary_reset,
            window_minutes=10080,
        )
        await usage_repo.add_entry(
            account_id=account_b.id,
            used_percent=20.0,
            window="primary",
            reset_at=primary_reset,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=account_b.id,
            used_percent=50.0,
            window="secondary",
            reset_at=secondary_reset,
            window_minutes=10080,
        )

        balancer = LoadBalancer(_repo_factory)
        selection = await balancer.select_account()

        assert selection.account is not None
        assert selection.account.id == account_b.id

        refreshed = await session.get(Account, account_a.id)
        assert refreshed is not None
        await session.refresh(refreshed)
        assert refreshed.status == AccountStatus.QUOTA_EXCEEDED


@pytest.mark.asyncio
async def test_load_balancer_reactivates_after_secondary_reset(db_setup):
    encryptor = TokenEncryptor()
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    primary_reset = now_epoch + 3600
    secondary_reset = now_epoch + 7200

    account = Account(
        id="acc_secondary_reset",
        email="secondary_reset@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access-reset"),
        refresh_token_encrypted=encryptor.encrypt("refresh-reset"),
        id_token_encrypted=encryptor.encrypt("id-reset"),
        last_refresh=now,
        status=AccountStatus.QUOTA_EXCEEDED,
        deactivation_reason=None,
    )

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)
        await accounts_repo.upsert(account)

        await usage_repo.add_entry(
            account_id=account.id,
            used_percent=5.0,
            window="primary",
            reset_at=primary_reset,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=account.id,
            used_percent=0.0,
            window="secondary",
            reset_at=secondary_reset,
            window_minutes=10080,
        )

        balancer = LoadBalancer(_repo_factory)
        selection = await balancer.select_account()

        assert selection.account is not None
        assert selection.account.id == account.id

        refreshed = await session.get(Account, account.id)
        assert refreshed is not None
        await session.refresh(refreshed)
        assert refreshed.status == AccountStatus.ACTIVE


@pytest.mark.asyncio
async def test_load_balancer_treats_weekly_only_primary_as_quota_window(db_setup):
    encryptor = TokenEncryptor()
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    weekly_reset = now_epoch + 7200
    plus_primary_reset = now_epoch + 3600
    plus_secondary_reset = now_epoch + 7200

    free_account = Account(
        id="acc_free_weekly_full",
        email="free_weekly_full@example.com",
        plan_type="free",
        access_token_encrypted=encryptor.encrypt("free-access"),
        refresh_token_encrypted=encryptor.encrypt("free-refresh"),
        id_token_encrypted=encryptor.encrypt("free-id"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )
    plus_account = Account(
        id="acc_plus_available",
        email="plus_available@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("plus-access"),
        refresh_token_encrypted=encryptor.encrypt("plus-refresh"),
        id_token_encrypted=encryptor.encrypt("plus-id"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)
        await accounts_repo.upsert(free_account)
        await accounts_repo.upsert(plus_account)

        await usage_repo.add_entry(
            account_id=free_account.id,
            used_percent=100.0,
            window="primary",
            reset_at=weekly_reset,
            window_minutes=10080,
        )
        await usage_repo.add_entry(
            account_id=plus_account.id,
            used_percent=20.0,
            window="primary",
            reset_at=plus_primary_reset,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=plus_account.id,
            used_percent=20.0,
            window="secondary",
            reset_at=plus_secondary_reset,
            window_minutes=10080,
        )

        balancer = LoadBalancer(_repo_factory)
        selection = await balancer.select_account()

        assert selection.account is not None
        assert selection.account.id == plus_account.id

        refreshed_free = await session.get(Account, free_account.id)
        assert refreshed_free is not None
        await session.refresh(refreshed_free)
        assert refreshed_free.status == AccountStatus.QUOTA_EXCEEDED


@pytest.mark.asyncio
async def test_load_balancer_select_account_uses_cached_rows_for_detached_accounts(db_setup):
    encryptor = TokenEncryptor()
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    account = Account(
        id="acc_detached_refresh",
        email="detached-refresh@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("access-detached"),
        refresh_token_encrypted=encryptor.encrypt("refresh-detached"),
        id_token_encrypted=encryptor.encrypt("id-detached"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)
        await accounts_repo.upsert(account)
        await usage_repo.add_entry(
            account_id=account.id,
            used_percent=10.0,
            window="primary",
            reset_at=now_epoch + 300,
            window_minutes=5,
        )

    balancer = LoadBalancer(_repo_factory)
    selection = await balancer.select_account()

    assert selection.account is not None
    assert selection.account.id == account.id
    assert selection.account.plan_type == "plus"


@pytest.mark.asyncio
async def test_load_balancer_prefers_newer_weekly_primary_over_stale_secondary(db_setup):
    encryptor = TokenEncryptor()
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())
    stale_reset = now_epoch + 1800
    weekly_reset = now_epoch + 7200
    plus_primary_reset = now_epoch + 3600
    plus_secondary_reset = now_epoch + 7200

    free_account = Account(
        id="acc_free_weekly_stale_secondary",
        email="free_weekly_stale_secondary@example.com",
        plan_type="free",
        access_token_encrypted=encryptor.encrypt("free-stale-access"),
        refresh_token_encrypted=encryptor.encrypt("free-stale-refresh"),
        id_token_encrypted=encryptor.encrypt("free-stale-id"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )
    plus_account = Account(
        id="acc_plus_weekly_control",
        email="plus_weekly_control@example.com",
        plan_type="plus",
        access_token_encrypted=encryptor.encrypt("plus-control-access"),
        refresh_token_encrypted=encryptor.encrypt("plus-control-refresh"),
        id_token_encrypted=encryptor.encrypt("plus-control-id"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)
        await accounts_repo.upsert(free_account)
        await accounts_repo.upsert(plus_account)

        await usage_repo.add_entry(
            account_id=free_account.id,
            used_percent=15.0,
            window="secondary",
            reset_at=stale_reset,
            window_minutes=10080,
            recorded_at=now - timedelta(days=2),
        )
        await usage_repo.add_entry(
            account_id=free_account.id,
            used_percent=100.0,
            window="primary",
            reset_at=weekly_reset,
            window_minutes=10080,
            recorded_at=now,
        )
        await usage_repo.add_entry(
            account_id=plus_account.id,
            used_percent=20.0,
            window="primary",
            reset_at=plus_primary_reset,
            window_minutes=300,
        )
        await usage_repo.add_entry(
            account_id=plus_account.id,
            used_percent=20.0,
            window="secondary",
            reset_at=plus_secondary_reset,
            window_minutes=10080,
        )

        balancer = LoadBalancer(_repo_factory)
        selection = await balancer.select_account()

        assert selection.account is not None
        assert selection.account.id == plus_account.id

        refreshed_free = await session.get(Account, free_account.id)
        assert refreshed_free is not None
        await session.refresh(refreshed_free)
        assert refreshed_free.status == AccountStatus.QUOTA_EXCEEDED


@pytest.mark.asyncio
async def test_load_balancer_filters_accounts_by_persisted_additional_usage(db_setup):
    encryptor = TokenEncryptor()
    now = utcnow()
    now_epoch = int(now.replace(tzinfo=timezone.utc).timestamp())

    exhausted_account = Account(
        id="acc_additional_full",
        email="additional_full@example.com",
        plan_type="pro",
        access_token_encrypted=encryptor.encrypt("access-full"),
        refresh_token_encrypted=encryptor.encrypt("refresh-full"),
        id_token_encrypted=encryptor.encrypt("id-full"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )
    eligible_account = Account(
        id="acc_additional_ok",
        email="additional_ok@example.com",
        plan_type="pro",
        access_token_encrypted=encryptor.encrypt("access-ok"),
        refresh_token_encrypted=encryptor.encrypt("refresh-ok"),
        id_token_encrypted=encryptor.encrypt("id-ok"),
        last_refresh=now,
        status=AccountStatus.ACTIVE,
        deactivation_reason=None,
    )

    async with SessionLocal() as session:
        accounts_repo = AccountsRepository(session)
        usage_repo = UsageRepository(session)
        additional_repo = AdditionalUsageRepository(session)
        await accounts_repo.upsert(exhausted_account)
        await accounts_repo.upsert(eligible_account)

        for account, used_percent in ((exhausted_account, 40.0), (eligible_account, 20.0)):
            await usage_repo.add_entry(
                account_id=account.id,
                used_percent=used_percent,
                window="primary",
                reset_at=now_epoch + 300,
                window_minutes=5,
                recorded_at=now,
            )

        await additional_repo.add_entry(
            account_id=exhausted_account.id,
            limit_name="codex_other",
            metered_feature="codex_bengalfox",
            window="primary",
            used_percent=100.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
            recorded_at=now,
        )
        await additional_repo.add_entry(
            account_id=eligible_account.id,
            limit_name="codex_other",
            metered_feature="codex_bengalfox",
            window="primary",
            used_percent=25.0,
            reset_at=now_epoch + 300,
            window_minutes=5,
            recorded_at=now,
        )

    balancer = LoadBalancer(_repo_factory)
    selection = await balancer.select_account(additional_limit_name="codex_spark")

    assert selection.account is not None
    assert selection.account.id == eligible_account.id



from __future__ import annotations

import ipaddress
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import Literal, Protocol

from app.modules.firewall.repository import FirewallRepositoryConflictError


class FirewallEntryLike(Protocol):
    ip_address: str
    created_at: datetime


class FirewallRepositoryPort(Protocol):
    async def list_entries(self) -> Sequence[FirewallEntryLike]: ...

    async def list_ip_addresses(self) -> set[str]: ...

    async def exists(self, ip_address: str) -> bool: ...

    async def add(self, ip_address: str) -> FirewallEntryLike: ...

    async def delete(self, ip_address: str) -> bool: ...


class FirewallValidationError(ValueError):
    pass


class FirewallIpAlreadyExistsError(ValueError):
    pass


@dataclass(frozen=True, slots=True)
class FirewallIpEntryData:
    ip_address: str
    created_at: datetime


@dataclass(frozen=True, slots=True)
class FirewallListData:
    mode: Literal["allow_all", "allowlist_active"]
    entries: list[FirewallIpEntryData]


class FirewallService:
    def __init__(self, repository: FirewallRepositoryPort) -> None:
        self._repository = repository

    async def list_ips(self) -> FirewallListData:
        rows = await self._repository.list_entries()
        entries = [FirewallIpEntryData(ip_address=row.ip_address, created_at=row.created_at) for row in rows]
        mode: Literal["allow_all", "allowlist_active"] = "allow_all" if not entries else "allowlist_active"
        return FirewallListData(mode=mode, entries=entries)

    async def add_ip(self, ip_address: str) -> FirewallIpEntryData:
        normalized = normalize_ip_address(ip_address)
        if await self._repository.exists(normalized):
            raise FirewallIpAlreadyExistsError("IP address already exists")
        try:
            row = await self._repository.add(normalized)
        except FirewallRepositoryConflictError as exc:
            raise FirewallIpAlreadyExistsError("IP address already exists") from exc
        return FirewallIpEntryData(ip_address=row.ip_address, created_at=row.created_at)

    async def remove_ip(self, ip_address: str) -> bool:
        normalized = normalize_ip_address(ip_address)
        return await self._repository.delete(normalized)

    async def is_ip_allowed(self, ip_address: str | None) -> bool:
        allowlist = await self._repository.list_ip_addresses()
        if not allowlist:
            return True
        if ip_address is None:
            return False
        try:
            normalized = normalize_ip_address(ip_address)
        except FirewallValidationError:
            return False
        return normalized in allowlist


def normalize_ip_address(value: str) -> str:
    raw = value.strip()
    if not raw:
        raise FirewallValidationError("IP address is required")
    try:
        return str(ipaddress.ip_address(raw))
    except ValueError as exc:
        raise FirewallValidationError("Invalid IP address") from exc



from __future__ import annotations

from datetime import datetime, timezone


def utcnow() -> datetime:
    return datetime.now(timezone.utc).replace(tzinfo=None)


def to_utc_naive(value: datetime) -> datetime:
    if value.tzinfo is None:
        return value
    return value.astimezone(timezone.utc).replace(tzinfo=None)


def naive_utc_to_epoch(dt: datetime) -> int:
    """Convert a naive-UTC datetime to Unix epoch seconds."""
    return int(dt.replace(tzinfo=timezone.utc).timestamp())


def from_epoch_seconds(value: int | float | None) -> datetime | None:
    if value is None:
        return None
    return datetime.fromtimestamp(value, tz=timezone.utc)



from __future__ import annotations

from app.core.errors import response_failed_event


def test_response_failed_event_includes_incomplete_details():
    event = response_failed_event("stream_incomplete", "Upstream closed stream", response_id="resp_1")

    response = event["response"]
    assert "incomplete_details" in response
    assert response["incomplete_details"] is None


def test_response_failed_event_accepts_incomplete_details():
    event = response_failed_event(
        "stream_incomplete",
        "Upstream closed stream",
        response_id="resp_1",
        incomplete_details={"reason": "max_output_tokens"},
    )

    response = event["response"]
    assert response.get("incomplete_details") == {"reason": "max_output_tokens"}



"""add transport to request_logs

Revision ID: 20260310_000000_add_request_logs_transport
Revises: 20260310_000000_fix_postgresql_enum_value_casing
Create Date: 2026-03-10
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260310_000000_add_request_logs_transport"
down_revision = "20260310_000000_fix_postgresql_enum_value_casing"
branch_labels = None
depends_on = None


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {column["name"] for column in inspector.get_columns(table_name)}


def upgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "request_logs")
    if not columns or "transport" in columns:
        return

    with op.batch_alter_table("request_logs") as batch_op:
        batch_op.add_column(sa.Column("transport", sa.String(), nullable=True))


def downgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "request_logs")
    if "transport" not in columns:
        return

    with op.batch_alter_table("request_logs") as batch_op:
        batch_op.drop_column("transport")
